SelectiveAcceptActivity subclass: #ConcurrentActivity
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Concurrent'!


!ConcurrentActivity methodsFor: 'activity setting'!

startSubProcessFor: aMessage
	"Create and start a new subprocess to consume a message."
	"The message to be consumed is passed as an argument.
	This allowes lazy creation of subprocesses when needed (when message exists)."

	[self subProcessBodyFor: aMessage] newProcess resume!

subProcessBodyFor: aMessage
	"Define the subprocess body."
	"Default body is to accept and consume the message."

	self subProcessAcceptMessage: aMessage! !

!ConcurrentActivity methodsFor: 'message handling'!

subProcessAcceptMessage: aMessage
	"Acceptance of a message by a subprocess."
	"Default is standard acceptMessage: as for a primary process.
	Note we use super in order to freeze the definition of accept (as in class Kernel2Activity).
	This is because method acceptMessage: will be redefined in subclasses
	in order to specify the behavior of the primary process."

	super acceptMessage: aMessage! !

ConcurrentActivity subclass: #SynchroConcurrentActivity
	instanceVariableNames: 'synchroMutualExclusion '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Concurrent'!


!SynchroConcurrentActivity methodsFor: 'initialize'!

privateInitialize
	"Initialize the semaphore for mutual exclusion of synchronization actions."

	super privateInitialize.
	synchroMutualExclusion := Semaphore forMutualExclusion! !

!SynchroConcurrentActivity methodsFor: 'accessing'!

synchroMutualExclusion
	^synchroMutualExclusion! !

!SynchroConcurrentActivity methodsFor: 'events'!

kernelEventAccept: aMessage
	"Dispatch onto synchronization events which should be mutually exclusive."

	super kernelEventAccept: aMessage.
	synchroMutualExclusion critical:
		[self synchroEventAccept: aMessage]!

kernelEventComplete: aMessage
	"Dispatch onto synchronization events which should be mutually exclusive."

	super kernelEventComplete: aMessage.
	synchroMutualExclusion critical:
		[self synchroEventComplete: aMessage]! !

!SynchroConcurrentActivity methodsFor: 'synchro events'!

synchroEventAccept: aMessage
	"This synchronization event method is called when the behavior accepts a message.
	It is used for synchronization purposes, for instance updating synchronization counters.
	At this abstract level, default behavior is empty."!

synchroEventComplete: aMessage
	"This synchronization event method is called once the behavior has completed computation of a message.
	It is used for synchronization purposes, for instance updating synchronization counters.
	At this abstract level, default behavior is empty."!

synchroEventReceive: aMessage
	"This synchronization event method is called once the behavior has received a message.
	It is used for synchronization purposes, for instance updating synchronization counters.
	At this abstract level, default behavior is empty."! !

!SynchroConcurrentActivity methodsFor: 'default classes'!

addressClass
	"Needs to be paired with class SynchroConcurrentAddress
	which defines atomic synchronization event for message reception."

	^SynchroConcurrentAddress! !

!SynchroConcurrentActivity methodsFor: 'compatibility constraints'!

addressConstraint
	"Enforce assumption defined in method addressClass."

	^SynchroConcurrentAddress! !

ImplicitReplyAddress subclass: #SynchroConcurrentAddress
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Concurrent'!


!SynchroConcurrentAddress methodsFor: 'message passing'!

asynchronousSend: aMessage inMessageQueue: queue
	"Ensure that asynchronous send (actual enqueueing of the message)
	AND the receive synchronization event are atomic."

	activity synchroMutualExclusion critical:
		[super asynchronousSend: aMessage inMessageQueue: queue.
		activity synchroEventReceive: aMessage]! !

!SynchroConcurrentAddress methodsFor: 'compatibility constraints'!

activityConstraint
	"SynchroConcurrentAddress method
		asynchronousSend:inMessageQueue:
			calls
	SynchroConcurrentActivity method
		synchroEventReceive:."

	^SynchroConcurrentActivity! !
