ImplicitReplyObject subclass: #GuardsObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Guards'!


!GuardsObject methodsFor: 'default classes'!

activityClass
	"Only purpose is to dispatch onto activity class GuardsActivity."

	^GuardsActivity! !

ImplicitReplyObject subclass: #ObjectGuardsObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Guards'!


!ObjectGuardsObject methodsFor: 'default classes'!

activityClass
	"Only purpose is to dispatch onto activity class ObjectGuardsActivity."

	^ObjectGuardsActivity! !

SelectiveAcceptActivity subclass: #Guards1Activity
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Guards'!


!Guards1Activity methodsFor: 'message handling'!

isAcceptableMessage: aMessage
	"First naive implementation of guards."	
	"Note that because of resending message which are not acceptable,
	it does not ensure message ordering preservation."

	^self satisfyGuardSelector: aMessage selector arguments: aMessage arguments! !

!Guards1Activity methodsFor: 'guard testing'!

findGuardSelector: selector
	"A guard is simply associated to a selector with the prefix guardOf.
	For instance the guard of method put: is the method guardOfPut:."

	^('guardOF' , selector) asSymbol!

satisfyGuardSelector: selector arguments: argumentsArray
	"Evaluate and return the guard method."

	^self perform: (self findGuardSelector: selector) withArguments: argumentsArray! !

Guards1Activity subclass: #GuardsActivity
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Guards'!


!GuardsActivity methodsFor: 'message handling'!

nextMessage
	"Look for the first message in the mailbox verifying its guard.
	If all current pending messages have false guards,
	then suspend until next received message and check its guard, etc...
	This makes use of method atIndex: which suspends until there is a ith message in the mailBox."

	"As opposed to the initial naive strategy of resending the message
	(class Guards1Activity), this strategy is more efficient
	and it also ensures message ordering preservation."

	"About guards reevaluation semantics:
	Guards are reevaluated (only) when looking for a new message to accept.
	There is no intra-object concurrency in this model of guards."

	| i message mailBox |
	mailBox := self mailBox.
	i := 1.
	message := mailBox atIndex: i.
	[self isAcceptableMessage: message] whileFalse:
		[i := i+1.
		message := mailBox atIndex: i].
	mailBox removeAtIndex: i.
	^message! !

GuardsActivity subclass: #ObjectGuardsActivity
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Guards'!


!ObjectGuardsActivity methodsFor: 'guard testing'!

satisfyGuardSelector: selector arguments: argumentsArray
	"Delegate the specification and evaluation of the guard to the active object."

	^oself perform: (self findGuardSelector: selector) withArguments: argumentsArray! !
